#!/usr/bin/python
# -*- coding: utf-8 -*-

''' Implements the Stack object.
'''

import copy


class Stack(object):
	''' Implements a stack object which stores compilation status.
	'''

	def __init__(self, value = None):
		''' Initializes the stack.
		'''
		if value is None:
			self._data = []
			self._symbols = []
		else:
			self._data = value
			self._symbols = [x.name for x in value]

	
	def __str__(self):
		return str(self._symbols)


	def __len__(self):
		return len(self._data)


	def __add__(self, data):
		result = Stack(copy.deepcopy(self._data))
		result += data

		return result


	def __iadd__(self, data):
		self._data += data
		self._symbols += [x.name for x in data]
		return self


	def __getitem__(self, key):
		return self._data[key]


	def __setitem__(self, key, value):
		self._data[key] = value
		self._symbols[key] = [x.name for x in value]

		return value	


	def __iter__(self):
		for i in self._data:
			yield i


	@property
	def symbols(self):
		''' Return a list of sumbols names in the stack.
		'''

		return list(self._symbols)


	def can_reduce(self, rule, eps_rule = None):
		''' Returns whether the given rule can be reduced.
		If the 1st symbol of the rule is epsilon, then
		self._data[-l].rule must be eps_rule.

		Each EPSILON is annotated which rule it matches.
		'''
		l = len(rule.right)
		if l > len(self._data):
			return False

		if rule.eps_action is not None: # This rule has an epsilon prepended
			try:
				if eps_rule is None or self._data[-l].rule != eps_rule:
					return False
			except AttributeError:
				return False

		return self._symbols[-l:] == rule.right


	def can_reduce_eps(self, rule):
		''' Returns whether the given rule can be reduced prepending epsilon
			into the stack.
		'''
		if rule.eps_action is None:
			return False

		l = len(rule.right) - 1
		if l > len(self._data) or l < 1:
			return False

		return self._symbols[-l:] == rule.right[1:]

