#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Implements some type checking using decorators.
'''

from types import *
from errors import *



EXTENDED_TYPES = {
	'NoneType': (NoneType),
	'ListType': (ListType),
	'DictType': (DictType),
	'DictionaryType': (DictType),
	'IntType' : (IntType),
	'LongType': (LongType),
	'StringTypes': (StringTypes), 
	'StringType': (StringType), 
	'UnicodeType': (UnicodeType),
	'FloatType': (FloatType), 
	'FunctionType': (FunctionType),
	'ClassType': (ClassType),
	'InstanceType': (InstanceType),
	'FileType' : (FileType),
	'TupleType' : (TupleType),
	
	# Complex types (list of strings, list of numbers, etc), always
	# use STRINGS in their values. E.g. use 'ListType', not ListType

	'NumberTypes': ('FloatType', 'IntType', 'LongType'),	# Any real number (not complex number) 
	'StringListType': {'ListType' : 'StringTypes'},				# A list of strings
	'NumberListType': {'ListType' : 'NumberTypes'},				# A list of numbers
	'StringOrNumberList': {'ListType': ('StringTypes', 'NumberTypes') },
	'NumberOrStringList': 'StringOrNumberList',
	'FunctionListType': {'ListType': ('NoneType', 'FunctionType')}
}




def check_types(typelist, instance, ExceptClass, msg):
	''' Check whether the instance is one of the types in list.
	Otherwise raises an instance of the ExceptClass(instance, msg) as an exception.
	'''
	if not isinstance(typelist, TupleType): # If not a List, convert it to a List
		typelist = (typelist, )

	for t in typelist:	# Check for every type, to see if one matches
		name = t
		if t in EXTENDED_TYPES.keys():	# If it's an extended type
			t = EXTENDED_TYPES[t]		# convert it to its value

		if isinstance(t, TupleType):
			for tt in t:
				try:
					check_types(tt, instance, ExceptClass, msg)
					return name		# Returns the matched type
				except ExceptClass:
					pass			# None of the extended types
		elif isinstance(t, DictType): # An iterable container where each element type is checked	
			try:
				q = check_types(tuple(t.keys()), instance, ExceptClass, msg) # The container type is checked
				if isinstance(t[q], ListType):
					return check_types(t[q], instance, ExceptClass, msg)

				for i in instance:			# Iterates the instance
					check_types(t[q], i, ExceptClass, msg)	# Checks for every element of the container
				return q

			except ExceptClass:
				pass
		elif isinstance(t, ListType): # An ordered list of types, element i-th must match type ith
			if len(t) != len(instance):
				raise ExceptClass(instance, msg)
		
				for tt, i in zip(t, instance):
					check_types(tt, i, ExceptClass, msg) 

			return name
		elif isinstance(instance, t):
			return name # Right type matched

	raise ExceptClass(instance, msg)  # Type not recognized.


